# typing
from __future__ import annotations
from collections.abc import Callable, Iterable

# ui
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.popup import Popup
from kivy.uix.widget import Widget
from kivy.uix.boxlayout import BoxLayout
from kivy.properties import ObjectProperty
from kivy.clock import Clock

# gamut
from .config import GAMUT_SESSION


class Modal(FloatLayout):
    def __init__(self, title: str, window_hint: tuple = (0.9, 0.9), **kwargs):
        self.popup = None
        self.title = title
        self.window_hint = window_hint
        super().__init__(**kwargs)

    def open(self) -> None:
        self.popup = Popup(title=self.get_popup_title(),
                           content=self,
                           auto_dismiss=False,
                           size_hint=self.window_hint)
        self.popup.open()

    def get_popup_title(self):
        return self.title

    def on_cancel(self) -> None:
        self.popup.dismiss()


class BaseDialog(Modal):
    def __init__(self,
                 title: str = 'LOAD FILE OR FOLDER',
                 **kwargs):
        self.last_path = GAMUT_SESSION.get('last_dir')
        super().__init__(title=title, **kwargs)

    def on_entry_added(self, chooser: Widget) -> None:
        if chooser.path != self.last_path:
            chooser.selection = []
            self.last_path = chooser.path
            self.popup.title = self.get_popup_title()

    def get_popup_title(self) -> None:
        return f"{self.title}: {self.last_path}"

    def sort_dirs(self, files: Iterable, filesystem: Iterable) -> Iterable:
        def key(x): return x.lower()
        return (sorted((f for f in files if filesystem.is_dir(f)), key=key) +
                sorted((f for f in files if not filesystem.is_dir(f)), key=key))


class LoadDialog(BaseDialog):
    def __init__(self,
                 on_load: Callable = None,
                 multiselect: bool = False,
                 dirselect: bool = False,
                 filters: Iterable = None,
                 **kwargs) -> None:
        self.on_load = on_load or (lambda path, selection: print(path, selection))
        self.multiselect = multiselect
        self.dirselect = dirselect
        self.filters = filters or []
        super().__init__(**kwargs)


class SaveDialog(BaseDialog):
    text_input = ObjectProperty(None)
    save_button = ObjectProperty(None)

    def __init__(self, on_save: Callable, title: str = 'LOAD FILE OR FOLDER', **kwargs) -> None:
        self.on_save = on_save  # must be a callable that takes two arguments: path, and filename
        super().__init__(title, **kwargs)
        Clock.schedule_once(lambda _: self.text_input.bind(text=self.on_text_change), 1)

    def on_text_change(self, instance: Widget, value: str) -> None:
        self.save_button.set_disabled(not value)


class SummaryField(BoxLayout):
    def __init__(self, key: str, value: str, **kwargs):
        self.key = " ".join(key.split(sep=None))
        self.value = " ".join(value.split(sep=None)) if value else "None"
        super().__init__(**kwargs)


class Summary(Modal):
    body = ObjectProperty(None)

    def __init__(self, summary: dict, **kwargs):
        super().__init__(window_hint=(0.5, 0.9), **kwargs)
        Clock.schedule_once(lambda _: self.__compile(summary))

    def __compile(self, summary: dict) -> None:
        for key in summary:
            self.body.add_widget(SummaryField(key=key.upper(), value=str(summary[key])))



# typing
from __future__ import annotations
from collections.abc import Iterable

# gui
from kivy.clock import Clock
from kivy.app import App
from kivy.uix.widget import Widget
from kivy.properties import ObjectProperty, StringProperty

# gamut
from .config import CORPUS_DIR, MOSAIC_DIR, CORPUS_CACHE, MOSAIC_CACHE, GAMUT_SESSION
from .utils import capture_exceptions, log_done, log_message, UserConfirmation
from .dialogs import LoadDialog, Summary
from .buttons import MenuItem

from ..features import Corpus, Mosaic
from ..config import AUDIO_FORMATS

# misc
import os


class MosaicFactoryWidget(Widget):
    """ Submodule for creating mosaics """

    create_mosaic_button = ObjectProperty(None)
    choose_target_button = ObjectProperty(None)
    mosaic_name = ObjectProperty(None)
    target = StringProperty('')

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.has_name = False
        Clock.schedule_once(lambda _: self.mosaic_name.bind(text=self.on_mosaic_name_update), 1)

    def get_selected_corpora(self) -> None:
        return [toggle.value for toggle in App.get_running_app().root.corpus_module.menu.corpora_menu.children if toggle.state == 'down']

    def update_mosaic_buttons(self) -> None:
        selected_corpora = self.get_selected_corpora()
        self.choose_target_button.set_disabled(not selected_corpora)
        value = all([self.has_name, selected_corpora, self.target])
        self.create_mosaic_button.set_disabled(not value)

    def on_mosaic_name_update(self, instance: Widget, value: str) -> None:
        self.has_name = bool(value)
        self.update_mosaic_buttons()

    def update_mosaic_menu(self) -> None:
        App.get_running_app().root.mosaic_module.menu.update_mosaic_menu()

    def load_target(self) -> None:
        """ Open load file dialog """
        def on_load(selected: Iterable):
            self.target = selected[0]
            GAMUT_SESSION.set('last_dir', os.path.dirname(self.target))
            self.update_mosaic_buttons()
        LoadDialog(on_load=on_load,
                   title='CHOOSE TARGET',
                   filters=[f"*{ft}" for ft in AUDIO_FORMATS]).open()

    @capture_exceptions
    @log_done
    def create_mosaic(self) -> None:
        """ Creates a mosaic based on selected corpus and chosen target """
        mosaic_name = self.mosaic_name.text
        log_message(f"Creating mosaic: {mosaic_name}...")
        corpus_names = self.get_selected_corpora()
        corpora = []
        for corpus_name in corpus_names:
            log_message(f"Loading corpus: {corpus_name}...")
            if corpus_name in CORPUS_CACHE:
                corpora.append(CORPUS_CACHE[corpus_name])
            else:
                path = os.path.join(CORPUS_DIR, f"{corpus_name}.gamut")
                corpus = Corpus().read(path)
                corpora.append(corpus)
                CORPUS_CACHE[corpus_name] = corpus
        log_message(f"Maching segments...")
        global MOSAIC_CACHE
        mosaic = Mosaic(target=self.target, corpus=corpora)
        mosaic.write(os.path.join(MOSAIC_DIR, f'{mosaic_name}.gamut'))
        MOSAIC_CACHE[mosaic_name] = mosaic
        self.update_mosaic_menu()


class MosaicMenuWidget(Widget):
    """ Submodule menu of existing mosaics """
    delete_mosaic_button = ObjectProperty(None)
    mosaic_summary_button = ObjectProperty(None)
    mosaic_menu = ObjectProperty(None)
    audio_module = ObjectProperty(None)

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.selected_mosaic = None
        Clock.schedule_once(lambda _: self.update_mosaic_menu(is_first_time=True), 1)

    def update_mosaic_menu(self, is_first_time: bool = False) -> None:
        self.clear_menu()
        # get full corpus paths
        paths = sorted([os.path.join(MOSAIC_DIR, basename) for basename in os.listdir(MOSAIC_DIR)])

        # get most recent file, if any
        most_recent = max(paths, key=os.path.getctime) if (paths and not is_first_time) else None

        # add a widget for each path, and select most recent by default
        for path in paths:
            mosaic_name = os.path.basename(os.path.splitext(path)[0])

            # set newest as selected
            if not is_first_time and path == most_recent:
                self.selected_mosaic = mosaic_name

            # create menu item toggle
            t = MenuItem(value=mosaic_name,
                         state='down' if path == most_recent else 'normal',
                         group='mosaics',
                         on_release=lambda toggle: self.on_select(toggle) or self.update_delete_button())

            self.mosaic_menu.add_widget(t)
        if not is_first_time:
            self.update_audio_synth_button()
        self.update_delete_button()

    def clear_menu(self) -> None:
        while self.mosaic_menu.children:
            self.mosaic_menu.clear_widgets()
        self.update_delete_button()

    def on_select(self, toggle: Widget) -> None:
        self.selected_mosaic = toggle.value if toggle.state == 'down' else None
        self.update_audio_synth_button()

    def delete_selected_mosaics(self) -> None:
        def on_confirm():
            for toggle in self.get_selected_toggles():
                self.mosaic_menu.remove_widget(toggle)
                os.remove(os.path.join(MOSAIC_DIR, f"{toggle.value}.gamut"))
            self.selected_mosaic = None
            self.update_delete_button()
            self.update_audio_synth_button()
            log_message(f"Mosaic deleted", log_type='error')
        UserConfirmation(on_confirm=on_confirm, long_text=f"You're about to delete this mosaic.").open()

    def update_audio_synth_button(self) -> None:
        App.get_running_app().root.mosaic_module.audio_module.synth_button.set_disabled(not bool(self.selected_mosaic))

    def get_selected_toggles(self) -> None:
        return [toggle for toggle in self.mosaic_menu.children if toggle.state == 'down']

    def update_delete_button(self) -> None:
        selected = self.get_selected_toggles()
        self.delete_mosaic_button.set_disabled(not bool(selected))
        self.mosaic_summary_button.set_disabled(len(selected) != 1)

    @capture_exceptions
    def open_summary(self):
        """ Opens a modal window with information about the currently selected mosaic """
        global MOSAIC_CACHE
        mosaic_name = self.get_selected_toggles()[0].value
        if mosaic_name in MOSAIC_CACHE:
            mosaic = MOSAIC_CACHE[mosaic_name]
        else:
            mosaic = Mosaic().read(os.path.join(MOSAIC_DIR, f'{mosaic_name}.gamut'))
            MOSAIC_CACHE[mosaic_name] = mosaic
        Summary(title="MOSAIC SUMMARY", summary=mosaic._summarize()).open()


class MosaicWidget(Widget):
    menu = ObjectProperty(None)
    factory = ObjectProperty(None)



# typing
from __future__ import annotations
from collections.abc import Iterable

# ui
from kivy.uix.widget import Widget
from kivy.properties import ObjectProperty
from kivy.clock import Clock
from kivy.app import App

# gamut
from .config import CORPUS_DIR, CORPUS_CACHE, GAMUT_SESSION
from .dialogs import LoadDialog, Summary
from .utils import log_message, capture_exceptions, log_done, UserConfirmation
from ..features import Corpus
from ..config import AUDIO_FORMATS
from .buttons import MenuItem

# misc
import os


def on_toggle_click(self, toggle: MenuItem, name: str) -> None:
    item_list = getattr(self, f'selected_{name}')
    delete_button = getattr(self, f'delete_{name}_button')
    if toggle.state == 'normal':
        item_list.remove(toggle.value)
    else:
        item_list.append(toggle.value)
    delete_button.set_disabled(not bool(item_list))


class CorpusFactoryWidget(Widget):
    sources_menu = ObjectProperty(None)
    features_menu = ObjectProperty(None)
    delete_sources_button = ObjectProperty(None)
    create_corpus_button = ObjectProperty(None)
    corpus_name = ObjectProperty(None)

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.sources = []
        self.selected_sources = []
        self.has_name = False
        self.selected_features = ['timbre']
        Clock.schedule_once(lambda _: self.corpus_name.bind(text=self.corpus_name_filter), 1)

    def update_create_corpus_button(self) -> None:
        self.create_corpus_button.set_disabled(not all([self.sources, self.selected_features, self.has_name]))

    def update_corpus_menu(self) -> None:
        App.get_running_app().root.corpus_module.menu.update_corpus_menu()

    def corpus_name_filter(self, instance: Widget, value: str) -> None:
        self.has_name = bool(value)
        self.update_create_corpus_button()

    def get_existent_corpus_names(self) -> Iterable:
        return [toggle.value.lower() for toggle in App.get_running_app().root.corpus_module.menu.corpora_menu.children]

    def add_sources(self) -> None:
        def on_load(selected):
            for source in selected:
                if source in self.sources or source.startswith('..'):
                    continue
                self.sources.append(source)
                self.sources_menu.add_widget(self.make_toggle(value=source, name='sources'))
            GAMUT_SESSION.set('last_dir', os.path.dirname(selected[0]))
            self.update_create_corpus_button()
        LoadDialog(multiselect=True, title="ADD SOURCES", dirselect=True, filters=[
                   f'*{ft}' for ft in AUDIO_FORMATS], on_load=on_load).open()

    def update_selected(self, name: str) -> None:
        setattr(self, f'selected_{name}', [toggle.value for toggle in getattr(
            self, f"{name}_menu").children if toggle.state == 'down'])
        btn_name = f'delete_{name}_button'
        if hasattr(self, btn_name):
            btn = getattr(self, btn_name)
            btn.set_disabled(not bool(getattr(self, f'selected_{name}')))
        self.update_create_corpus_button()

    def delete_selected(self, name: str) -> None:
        selected = getattr(self, f'selected_{name}')
        menu = getattr(self, f'{name}_menu')
        for item in selected:
            if name == 'sources':
                self.sources.remove(item)
            for toggle in menu.children:
                if toggle.value == item:
                    menu.remove_widget(toggle)
                    break
        setattr(self, f'selected_{name}', [])
        getattr(self, f'delete_{name}_button').set_disabled(True)
        self.update_create_corpus_button()

    @capture_exceptions
    @log_done
    def create_corpus(self) -> None:
        corpus_name = self.corpus_name.text.strip()
        log_message(f'Creating corpus: {corpus_name}...')
        global CORPUS_CACHE
        corpus = Corpus(source=self.sources, features=self.selected_features)
        corpus.write(os.path.join(CORPUS_DIR, f'{corpus_name}.gamut'))
        CORPUS_CACHE[corpus_name] = corpus
        self.update_corpus_menu()

    def make_toggle(self, value: str, name: str) -> None:
        return MenuItem(value=value, on_release=lambda _: self.update_selected(name))


class CorpusMenuWidget(Widget):
    corpora_menu = ObjectProperty(None)
    selected_corpora = ObjectProperty(None)
    delete_corpora_button = ObjectProperty(None)
    corpus_summary_button = ObjectProperty(None)

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        Clock.schedule_once(lambda _: self.update_corpus_menu(is_first_time=True), 1)

    def update_corpus_menu(self, is_first_time: bool = False) -> None:
        self.clear_menu()
        # get full corpus paths
        paths = sorted([os.path.join(CORPUS_DIR, basename) for basename in os.listdir(CORPUS_DIR)])

        # get most recent file, if any
        most_recent = max(paths, key=os.path.getctime) if (paths and not is_first_time) else None

        # add a widget for each path, and select most recent by default
        for path in paths:
            corpus_name = os.path.basename(os.path.splitext(path)[0])

            t = MenuItem(value=corpus_name,
                         state='down' if path == most_recent else 'normal',
                         on_release=lambda _: self.update_delete_button() or self.update_mosaic_buttons())
            self.corpora_menu.add_widget(t)
        if not is_first_time:
            #
            self.update_mosaic_buttons()
        self.update_delete_button()

    def get_selected_toggles(self) -> Iterable:
        return [toggle for toggle in self.corpora_menu.children if toggle.state == 'down']

    def update_mosaic_buttons(self) -> None:
        App.get_running_app().root.mosaic_module.factory.update_mosaic_buttons()

    def update_delete_button(self) -> None:
        selected = self.get_selected_toggles()
        self.delete_corpora_button.set_disabled(not selected)
        self.corpus_summary_button.set_disabled(not len(selected) == 1)

    @capture_exceptions
    def open_summary(self):
        """ Opens a modal window with information about the currently selected corpus """
        global CORPUS_CACHE
        corpus_name = self.get_selected_toggles()[0].value
        if corpus_name in CORPUS_CACHE:
            corpus = CORPUS_CACHE[corpus_name]
        else:
            corpus = Corpus().read(os.path.join(CORPUS_DIR, f'{corpus_name}.gamut'))
            CORPUS_CACHE[corpus_name] = corpus
        Summary(title="CORPUS SUMMARY", summary=corpus._summarize()).open()

    def delete_selected_corpora(self) -> None:
        selected = self.get_selected_toggles()
        num_selected = len(selected)
        item_name = 'corpus' if num_selected == 1 else 'corpora'
        items = f'this {item_name}' if num_selected == 1 else f'these {num_selected} {item_name}'

        def on_confirm():
            for toggle in selected:
                self.corpora_menu.remove_widget(toggle)
                os.remove(os.path.join(CORPUS_DIR, f"{toggle.value}.gamut"))
            self.update_delete_button()
            log_message(f"{num_selected} {item_name} deleted", log_type='error')
        UserConfirmation(on_confirm=on_confirm, long_text=f"You're about to delete {items}").open()

    def clear_menu(self) -> None:
        while self.corpora_menu.children:
            self.corpora_menu.clear_widgets()
        self.update_delete_button()


class CorpusWidget(Widget):
    menu = ObjectProperty(None)
    factory = ObjectProperty(None)



from __future__ import annotations
from scipy.signal import get_window
from collections.abc import Iterable
import numpy as np

from .config import CONSOLE, ENVELOPE_TYPES
from math import ceil
from typing_extensions import Self


class Points(np.ndarray):
    """ 
    A subclass of a numpy array, mostly to allow inline method concatenation — e.g. x.a().b().c()
    """

    def __array_finalize__(self, obj) -> Self:
        return obj

    def __new__(cls, array: np.ndarray | list = []) -> Self:
        return np.asarray(array).view(cls)

    def fill(self, shape: Iterable, value: int | float) -> Self:
        tmp = np.empty(shape=shape)
        tmp.fill(value)
        return Points(tmp)

    def concat(self, x: Points | np.ndarray, prepend: bool = False) -> Points:
        return Points(np.concatenate([x, self] if prepend else [self, x]))

    def resample(self, N: int) -> Points:
        if len(self) == N:
            return self
        return Points(np.interp(np.linspace(0, len(self) - 1, N), np.arange(0, len(self)), self))

    def scale(self, out_min: float | int = 0.0, out_max: float | int = 1.0) -> Points:
        amin, amax = np.amin(self), np.amax(self)
        return Points(((self - amin) / (amax - amin)) * (out_max - out_min) + out_min)

    def replicate(self, n: int, axis: int = 0) -> Points:
        return Points(np.repeat(self, repeats=n, axis=axis))

    def wrap(self, n: int = 1) -> Self:
        out = self
        for _ in range(n):
            out = Points([out])
        return out

    def quantize(self, n: float | int = 1) -> Self:
        return Points(np.round(self / n) * n)

    def abs(self) -> Self:
        return Points(np.abs(self))

    def clip(self, min: float | int | None = None, max: float | int | None = None) -> Self:
        if min:
            self[self < min] = min
        if max:
            self[self > max] = max
        return self


class Envelope:
    """
    ``Envelope`` instances are meant to be used as dynamic audio control parameters, when calling the ``Mosaic.to_audio()`` method.
    For instance, to temporally change audio parameters such as grain duration, panning, playback speed, etc.
    """

    def __init__(self, shape: Iterable | str = "cosine") -> None:
        self.type = None
        self._points = None
        if isinstance(shape, Iterable) and type(shape) != str:
            self.__clean_points(shape)
        elif isinstance(shape, str):
            self.type = shape
        else:
            CONSOLE.error(
                TypeError,
                'Envelope must be initialized with either a window type (e.g. "hann") or an iterable of envelope points')

    def __str__(self) -> str:
        return f"<{self.__class__.__name__.capitalize()}: {self.type if self.type else str(self._points)}>"

    def __validate_points(self, points: Iterable) -> Points:
        if isinstance(points[0], Iterable):
            prev_idx = points[0][0]
            for point in points[1:]:
                if point[0] <= prev_idx:
                    CONSOLE.error(ValueError, f'Point indices must be in ascending order: {points}')
                prev_idx = point[0]
            return Points(points)
        return Points(list(enumerate(points)))

    def __clean_points(self, points: Iterable) -> None:
        pts = self.__validate_points(points)
        y = Points()
        for i in range(len(pts) - 1):
            segment = pts[i:i+2, 1]
            size = int(pts[i+1, 0] - pts[i, 0] + 1)
            y = y.concat(segment.resample(size)[:-1])
        self._points = y.concat([pts[-1][1]])

    def print_shape_choices(self) -> None:
        """ Prints all supported envelope shape types """
        print(f'shape choices: {ENVELOPE_TYPES}')

    def view(self, grid: bool = True) -> None:
        """ Helper method to visualize the shape of the envelope """
        import matplotlib.pyplot as plt
        y = get_window(self.type, Nx=64) if self.type else self._points
        plt.plot(y)
        plt.title(f"Envelope{f' ({self.type})' if self.type else ''}")
        plt.xlabel('time')
        plt.ylabel('control value')
        if self.type:
            plt.xticks([])
        if grid:
            plt.grid()
        plt.show()

    @property
    def points(self) -> Points:
        return self._points

    def get_points(self, N: int) -> Points:
        if self.type:
            return Points(get_window(self.type, N))
        return self._points.resample(N)


def object_to_points(param, N: int) -> Points:
    """ resolve parameter into a ``Points`` instance based on type """
    if isinstance(param, Envelope):
        return param.get_points(N)
    elif isinstance(param, Iterable):
        return Envelope(shape=param).get_points(N)
    else:
        return Points().fill(N, param)


def plot_envelope_list(env_list: Iterable, rows: int = 1) -> None:
    from matplotlib import pyplot as plt

    envs = np.array([object_to_points(x, 100) for x in env_list]).T
    cumsum = envs.sum(axis=1)[:, np.newaxis]
    envs = (envs / cumsum).T
    plt.figure(1)

    for i in range(len(env_list)):
        plt.subplot(int(f'{rows}{max(1, ceil(len(env_list)/rows))}{i+1}'),
                    ylim=[0, 1],
                    xlabel='time',
                    ylabel='value',
                    xticks=[],
                    yticks=np.round(np.linspace(0, 1, 10), 1),
                    title=f'Envelope {i + 1}'
                    )
        plt.plot(envs[i], color=['dodgerblue', 'tomato', 'coral', 'salmon'][i % 4],)
    plt.tight_layout()
    plt.show()



# typing
from __future__ import annotations

# kivy
from kivy.app import App
from kivy.uix.widget import Widget
from kivy.properties import ObjectProperty, StringProperty

# gamut
from .config import MOSAIC_CACHE, MOSAIC_DIR, GAMUT_SESSION
from .utils import parse_param_string, capture_exceptions, log_done, log_message
from .dialogs import SaveDialog
from ..features import Mosaic

# misc
import os


class Param(Widget):
    """
    Audio parameter widget
    """
    label = StringProperty('')
    value = StringProperty('')

    def list_filter(self, value: str, from_undo: bool = False) -> str | None:
        """ Text input filter for list-compatible parameters """
        if value.isdigit() or value in '.() ':
            return value

    def window_filter(self, value: str, from_undo: bool = False) -> str | None:
        """ Text input filter for list- and string-compatible parameters """
        if self.list_filter(value, from_undo) or value.isalpha():
            return value


class AudioWidget(Widget):
    """
    Audio module widget
    """
    params = ObjectProperty(None)
    synth_button = ObjectProperty(None)
    save_button = ObjectProperty(None)
    play_button = ObjectProperty(None)
    stop_button = ObjectProperty(None)

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.audio_buffer = None

    def get_selected_mosaic(self) -> Widget:
        """ Short hand method to access selected mosaic in MosaicWidget """
        return App.get_running_app().root.mosaic_module.menu.selected_mosaic

    @capture_exceptions
    def get_parsed_params(self):
        """ Parse input audio parameters """
        params = {}
        for param in self.params.children:
            value = parse_param_string(param.text_input.text)
            params[param.key] = value
        return params

    @capture_exceptions
    @log_done
    def synth_audio(self) -> None:
        """ Triggers audio mosaic synthesis """
        self.stop_audio()
        mosaic_name = self.get_selected_mosaic()
        log_message(f'Synthesizing mosaic: {mosaic_name}...')
        params = self.get_parsed_params()
        if mosaic_name in MOSAIC_CACHE:
            mosaic = MOSAIC_CACHE[mosaic_name]
        else:
            path = os.path.join(MOSAIC_DIR, f"{mosaic_name}.gamut")
            mosaic = Mosaic().read(path)
        self.audio_buffer = mosaic.to_audio(**params)
        self.update_play_and_save_buttons()

    def update_play_and_save_buttons(self) -> None:
        """ Updates disabled state of buttons based on presence of audio buffer """
        val = not bool(self.audio_buffer)
        self.play_button.set_disabled(val)
        self.save_button.set_disabled(val)
        self.stop_button.set_disabled(val)

    def play_audio(self) -> None:
        """ Writes audio file at chosen directory """
        self.audio_buffer.play(blocking=False)

    def stop_audio(self) -> None:
        """ Stops audio buffer playback """
        if self.audio_buffer:
            self.audio_buffer.stop()

    @capture_exceptions
    @log_done
    def save_audio(self) -> None:
        """ Writes audio file at chosen directory """
        def on_save(path, text):
            GAMUT_SESSION.set('last_dir', os.path.dirname(path))
            filename = os.path.join(path, text)
            filename = f"{os.path.splitext(filename)[0]}.wav"
            log_message(f'Saving audio file: {filename}')
            self.audio_buffer.write(filename)
        SaveDialog(title="SAVE AUDIO MOSAIC", on_save=on_save).open()



from __future__ import annotations
import numpy as np
from typing import Any, Callable
from collections.abc import Iterable


def get_nested_value(obj: object, path: str | None, separator: str = ".") -> Any:
    """
    Helper function to get an a nested value from an ``object`` of aribirary depth

    obj: object
        Object to extract value from.

    path: str | None
        path to traverse ``obj``, separated by ``separator``.

    separator: str
        substring by which to parse ``path``
    """
    output = obj
    for key in path.split(separator):
        t = type(output)
        if t == list:
            output = output[int(key)]
        elif t == dict:
            output = output[key]
        else:
            output = getattr(output, key)
    return output


def __set_nested_value(obj: object, paths: Iterable, replace: object) -> Any:
    if not paths:
        obj = replace
        return obj
    k = paths[0]
    if isinstance(obj, list) or isinstance(obj, np.ndarray):
        obj[int(k)] = __set_nested_value(obj[int(k)], paths[1:], replace)
    elif isinstance(obj, dict):
        obj[k] = __set_nested_value(obj[k], paths[1:], replace)
    else:
        setattr(obj, k, __set_nested_value(getattr(obj, k), paths[1:], replace))
    return obj


def set_nested_value(obj: object, path: str, replace: object, separator: str = '.') -> Any:
    """
    Helper function to update an a nested value from an ``object`` of aribirary depth

    obj: object
        Object to extract value from.

    replace:

    path: str | None
        path to traverse ``obj``, separated by ``separator``.

    separator: str
        substring by which to parse ``path``
    """
    return __set_nested_value(obj, path.split(separator), replace)


def resample_array(array: np.ndarray, N: int) -> np.ndarray:
    return np.interp(np.linspace(0, len(array) - 1, N), np.arange(0, len(array)), array)


def catch_keyboard_interrupt(interrupt_func: Callable | None = None) -> Callable:
    def decorator(function):
        def wrapper(*args, **kwargs):
            try:
                function(*args, **kwargs)
            except KeyboardInterrupt:
                print()
                if interrupt_func:
                    return interrupt_func()
        return wrapper
    return decorator



from __future__ import annotations
# librosa
from librosa import magphase, stft, samples_like, load
from librosa.feature import mfcc, chroma_stft, rms, zero_crossing_rate
from librosa.beat import tempo

# gamut
from .controls import Points, Envelope, object_to_points
from .utils import resample_array
from .audio import AudioBuffer
from .config import FILE_EXT, CONSOLE, ANALYSIS_TYPES, MIME_TYPES, AUDIO_DIR, get_elapsed_time
from .data import KDTree

# os
from os.path import realpath, basename, isdir, splitext, join, commonprefix, relpath, dirname
from os import walk, rename

# misc utils
import filetype
from random import choices, random
from copy import deepcopy
import datetime
import os

# typing
from typing_extensions import Self
from collections.abc import Iterable
from abc import ABC, abstractmethod

# numpy
import numpy as np


class Analyzer(ABC):
    """ 
    Abstract base class from which the ``Corpus`` and ``Mosaic`` classes inherit. 
    It provides the base methods for feature extraction and reading/writing ``.gamut`` files.

    n_mfcc: int = 13
        Number of mel frequency cepstral coefficients to use in mfcc analysis.

    hop_length: int = 512
        Size in audio samples of space between windowing frames.

    win_length: int = 1024
        Size in audio samples of windowing frames.

    n_fft: int = 1024
        Number of FFT bins
    """

    def __init__(self,
                 n_mfcc: int = 13,
                 hop_length: int = 512,
                 win_length: int = 1024,
                 n_fft: int = 1024) -> None:
        self.n_mfcc = n_mfcc
        self.hop_length = hop_length
        self.n_fft = n_fft
        self.win_length = win_length
        self.type = self.__get_type()
        self.name = None
        self.portable = False

    @abstractmethod
    def _serialize(self):
        raise NotImplementedError

    @abstractmethod
    def _preload(self):
        raise NotImplementedError

    @abstractmethod
    def _summarize(self):
        raise NotImplementedError

    def summarize(self) -> None:
        """ Prints a summary of the current structure of the object """
        summary = self._summarize()
        if self.name:
            summary = {'name': self.name, **summary}
        line = "".join("-" for _ in range(90))
        CONSOLE.log_process(f"*** {self.type.upper()} SUMMARY ***\n{CONSOLE.c3}{line}").print()
        for key in summary:
            val = summary[key]
            print(f"{CONSOLE.c4}{key.upper()}: {CONSOLE.reset}{val}")
        print(f'{CONSOLE.c3}{line}{CONSOLE.reset}')

    @get_elapsed_time
    def write(self, output: str, portable: bool = False) -> None:
        """ Writes a ``.gamut`` file to disk """
        self.portable = portable
        self.name = splitext(basename(output))[0]
        output_dir = splitext(output)[0]
        CONSOLE.log_disk_op(f'{"" if portable else "non-"}portable {self.type}', f'{realpath(output_dir)}{FILE_EXT}').print()
        serialized_object = self._serialize()

        # write file and set correct file extension
        np.save(output_dir, serialized_object)
        rename(output_dir + '.npy', output_dir+FILE_EXT)
        return self

    @get_elapsed_time
    def read(self, file: str, warn_user=False) -> Self:
        """ Reads a ``.gamut`` file from disk """
        if warn_user:
            CONSOLE.warn(f"This {self.type} already has a source")

        # validate file
        if splitext(file)[1] != FILE_EXT:
            CONSOLE.error(ValueError, 'Wrong file extension. Provide a directory for a {} file'.format(FILE_EXT))

        serialized_object = np.load(file, allow_pickle=True).item()
        if serialized_object['type'] != self.type:
            CONSOLE.error(TypeError, 'The specified file .gamut file is a {}, not a {}.'.format(
                serialized_object['type'], self.type))
        is_portable = serialized_object['portable']
        CONSOLE.log_disk_op(f'{"" if is_portable else "non-"}portable {self.type}', basename(file), read=True).print()

        serialized_object = self._preload(serialized_object)

        # assign attributes to self
        for attr in serialized_object:
            if hasattr(self, attr):
                setattr(self, attr, serialized_object[attr])

        return self

    def __get_type(self) -> str:
        """ Helper function to get subclass name """
        return self.__class__.__name__.lower()

    def _analyze_audio_file(self, y: np.ndarray, features: Iterable, sr: int | None = None) -> tuple:
        """ Extracts audio features from an ``ndarray`` of audio samples """
        S = magphase(stft(y=y,
                          n_fft=self.n_fft,
                          win_length=self.win_length,
                          hop_length=self.hop_length))[0]

        analysis = []
        if 'timbre' in features:
            mfcc_features = mfcc(S=S,
                                 sr=sr,
                                 n_mfcc=self.n_mfcc,
                                 hop_length=self.hop_length)

            analysis.extend(mfcc_features)

        if 'pitch' in features:
            zerox = zero_crossing_rate(y=y,
                                       frame_length=self.win_length,
                                       hop_length=self.hop_length)
            loudness = rms(S=S,
                           frame_length=self.win_length,
                           hop_length=self.hop_length)
            chroma_features = chroma_stft(S=S,
                                          sr=sr,
                                          n_fft=self.n_fft,
                                          win_length=self.win_length,
                                          hop_length=self.hop_length)
            analysis.extend(np.concatenate([chroma_features, loudness, zerox]))

        analysis = np.array(analysis).T[:-1]

        markers = samples_like(X=analysis,
                               hop_length=self.hop_length,
                               n_fft=self.n_fft,
                               axis=0)

        return analysis, markers


class Corpus(Analyzer):
    """ 
    A ``Corpus`` represents a collection of one or more audio sources, from which a ``Mosaic`` can be built.
    Internally, the audio sources are analyzed and decomposed into grains.\n
    Based on the audio features of these grains (e.g., timbre or pitch content), a k-dimensional search tree is built, 
    which helps to optimize the process of finding the best matches for a given audio target.

    source: str | list | None = None
        Source file(s) from which to build the ``Corpus``. ``source`` can be either a ``str`` or a list of ``str``, where 
        ``str`` is an audio file path or a directory of audio files. Note that, for any directory path, ``Corpus`` will recursively look for
        any ``.wav``, ``.aif``, or ``.mp3`` audio files in folders and subfolders.

    max_duration: str | list | None = None
        Maximum audio file duration to use in ``Corpus``. Applies to all audio files found in ``source``.

    n_mfcc: int = 13
        Number of mel frequency cepstral coefficients to use in audio analysis.

    hop_length: int = 512
        hop size in audio samples.

    win_length: int = 1024
        window size in audio samples.

    n_fft: int = 512
        Number of FFT bins

    features: Iterable = ['timbre']
        List of ``str``, specifying audio feature(s) to use for audio source analysis. The options are ``"timbre"`` and/or ``"pitch"``.

    leaf_size: int = 10
        Maximum number of data items per leaf in the k-dimensional binary search tree.
    """

    def __init__(self,
                 source: str | list | None = None,
                 max_duration: int | None = None,
                 leaf_size: int | None = 15,
                 features: Iterable = ['timbre'],
                 *args,
                 **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.source = source
        self.max_duration = max_duration
        self.features = features
        self.total_duration = 0
        self.source_root = ""
        self.tree = KDTree(leaf_size=leaf_size)
        CONSOLE.reset_counter('Analyzing audio samples: ')

        for f in self.features:
            if f in ANALYSIS_TYPES:
                continue
            CONSOLE.error(
                ValueError,
                f'"{f}" is not a valid feature option. To see the list of valid features, use the {self.type.capitalize()}.print_feature_choices() class method.')
        if len(self.features) == 0:
            CONSOLE.error(
                ValueError, f'You must specify at least one audio feature to instantiate a {self.type.capitalize()} instance')

        self.soundfiles = []
        if self.source:
            self.source = [self.source] if isinstance(self.source, str) else self.source
            self.source.append(os.path.join(AUDIO_DIR, 'silence.mp3'))
            self.__build()

    @get_elapsed_time
    def __build(self) -> None:
        """ build corpus from `source` """
        CONSOLE.log_process('\N{brain} Building audio corpus...').print()
        data = self.__compile(source=None, data=[], excluded_files=[])
        CONSOLE.counter.finish()
        self.__set_source_root()
        self.tree.build(data=data, vector_path='features')

    def _summarize(self) -> dict:
        filenames = "\n"
        for i, sf in enumerate(self.soundfiles):
            if i == -1:
                continue
            fn = basename(sf['file'])
            filenames += [f'\t{fn}', f', {fn}', f', {fn}\n'][[0, 1, 1, 2][i % 4]]

        return {
            "source root": self.source_root,
            "portable": self.portable,
            "duration (H:M:S)": str(datetime.timedelta(seconds=int(self.total_duration))),
            "max. duration per source": f'{self.max_duration}' + ("s" if self.max_duration else ""),
            "max. tree leaf size": self.tree.leaf_size,
            "analysis features": ", ".join(self.features),
            f'sources ({len(self.soundfiles)})': filenames,
        }

    def __compile(self, source: list | str | None = None, excluded_files: list = [], data: list = []) -> None:
        """ recursively collects and extracts features from all audio files in `source` """
        source = source or self.source
        source_type = type(source or self.source)
        if source_type == list:
            for x in source:
                self.__compile(x, excluded_files=excluded_files, data=data)
            return data

        source_path = realpath(source)
        if isdir(source_path):
            for root, _, files in walk(source_path):
                for f in files:
                    self.__compile(join(root, f), excluded_files=excluded_files, data=data)
            return data

        filename = splitext(basename(source))[0]
        kind = filetype.guess(source_path)
        if kind is None or kind.mime not in MIME_TYPES or filename in excluded_files:
            return data

        excluded_files.append(filename)

        y, sr = load(path=source, sr=None, mono=True,
                     duration=self.max_duration)
        source_id = len(self.soundfiles)
        self.soundfiles.append({
            'file': source,
            'sr': sr,
            'y': y
        })
        self.total_duration += len(y)/sr
        analysis, markers = self._analyze_audio_file(y=y, features=self.features, sr=sr)
        CONSOLE.counter.next()
        data.extend([{
            'source': source_id,
            'marker': marker,
            'features': features,
        } for features, marker in zip(analysis, markers)])

        return data

    @classmethod
    def print_feature_choices(cls) -> None:
        print(f'feature choices: {ANALYSIS_TYPES}')

    def __set_source_root(self) -> None:
        """ remove commonprefix from file paths to avoid size overhead when writing corpus to disk """

        if not self.soundfiles:
            CONSOLE.error(ValueError, 'Corpus does not contain any source audio files')
        if len(self.soundfiles) == 1:
            return
        self.source_root = dirname(commonprefix([sf['file'] for sf in self.soundfiles]))
        for i, sf in enumerate(self.soundfiles):
            self.soundfiles[i]['file'] = relpath(sf['file'], self.source_root)

    def _serialize(self) -> dict:
        """ called from within write method """
        corpus = deepcopy({**vars(self), 'tree': vars(self.tree)})

        # if not portable, delete audio samples from soundfiles on write
        if not self.portable:
            for sf in corpus['soundfiles']:
                del sf['y']
        return corpus

    def _preload(self, obj: object) -> dict:
        """ called from within read method """
        gamut_type = obj['type']
        if gamut_type != self.type:
            CONSOLE.error(ValueError, f'source file should be a {self.type}, not a {gamut_type}')
        tree = KDTree()
        tree.read(obj['tree'])
        obj['tree'] = tree

        # re-load audio files if corpus file is not portable
        if not obj['portable']:
            CONSOLE.counter.message = CONSOLE.log_subprocess('Loading audio files: ')
            for sf in obj['soundfiles']:
                path = join(obj['source_root'], sf['file'])
                sf['y'] = load(path, sr=sf['sr'])[0]
                CONSOLE.counter.next()
            CONSOLE.counter.finish()
        return obj

    def read(self, file: str) -> None:
        return super().read(file, warn_user=self.source)


class Mosaic(Analyzer):
    """ 
    A ``Mosaic`` represents the blueprint from which an audio mosaic can be synthesized.
    Given an input audio target file path and a ``Corpus`` instance, a virtual representation of 
    an audio mosaic is built, to be later converted to audio with the ``to_audio()`` method.\n

    By "blueprint" it's meant that a ``Mosaic`` instance generates and stores the necessary
    information to synthesize an audio mosaic, but does not automatically generate the audio mosaic.\n

    This allows the user to create different versions, based on the audio parameters passed to the 
    ``to_audio()`` method.

    target: str | None = None
        File path to target audio file.

    corpus: Iterable | Corpus = None
        ``Corpus`` instance to reconstruct ``target``

    sr: int | None = None
        Sampling rate of output audio file

    beat_unit: float | int | None = None
        Optional argument to set the grain rate to a beat unit relative to detected tempo (e.g., 1/4, 1/8, 1/16, etc.).
        Works best when ``target`` has a steady and perceptible tempo.
    """

    def __init__(self,
                 target: str | None = None,
                 corpus: Iterable | Corpus | None = None,
                 sr: int | None = None,
                 beat_unit: float | int | None = None,
                 *args,
                 **kwargs) -> None:
        self.__validate(target, corpus)
        super().__init__(*args, **kwargs)

        self.target = target
        self.sr = sr
        self.frames = []
        self.beat_unit = beat_unit
        self.features = []
        self.duration = None

        corpora = self.__parse_corpus(corpus, [])
        self.soundfiles = {i: {} for i in range(-1, len(corpora or []))}

        if self.target and corpora:
            self.features = corpora[0].features
            self.__build(corpora=corpora, sr=sr)

    def __validate(self, target: str | None, corpus: Iterable | Corpus | None) -> None:
        if any([target, corpus]) and not all([target, corpus]):
            CONSOLE.error(
                ValueError,
                f'You must either provide both target and corpus attributes, or leave them blank to build Mosaic from {FILE_EXT} file')
        if not target:
            return
        target_path = realpath(target)
        if isdir(target_path):
            CONSOLE.error(ValueError, f'{target} is not a valid audio file path')
        kind = filetype.guess(target_path)
        if kind is None or kind.mime not in MIME_TYPES:
            CONSOLE.error(ValueError, f'{basename(target)} seems to be an invalid or unsupported audio file format.')

    def __parse_corpus(self, corpus: Iterable | Corpus, corpora: Iterable) -> Iterable:
        if not corpus:
            return

        if isinstance(corpus, Iterable):
            # check that corpus features are compatible
            prev_features = set(corpus[0].features)
            for i, c in enumerate(corpus[1:], 1):
                current_features = set(c.features)
                if prev_features != current_features:
                    CONSOLE.error(
                        ValueError,
                        f'Corpus at index {i} has a different set of features ({", ".join(c.features)}) than corpus at index {i-1} ({", ".join(corpus[i-1].features)}). When using more than one corpus, make sure they are based on the same feature set.')
                prev_features = current_features
            # recursively unpack corpora
            for c in corpus:
                self.__parse_corpus(c, corpora=corpora)
        elif isinstance(corpus, Corpus):
            corpora.append(corpus)
        else:
            CONSOLE.error(ValueError, f'{corpus} is not a corpus')
        return corpora

    @get_elapsed_time
    def __build(self, corpora: Iterable, sr: int | None = None) -> None:
        num_corpora = len(corpora)
        CONSOLE.log_process(
            f'\N{brain} Building mosaic for {basename(self.target)} from {"corpus" if num_corpora == 1 else f"{num_corpora} corpora"}...').print()
        CONSOLE.log_subprocess('Loading target...').print()
        y, self.sr = load(self.target, sr=sr)

        self.duration = len(y) / self.sr

        if self.beat_unit:
            self.hop_length = int((self.sr * 60) / (tempo(y=y, sr=self.sr)[0] / self.beat_unit))
        target_analysis = self._analyze_audio_file(y=y, features=corpora[0].features, sr=self.sr)[0]

        # include separate corpus for target
        self.soundfiles[-1] = {
            'source_root': "",
            'max_duration': None,
            'sources': {
                0: {
                    'file': self.target,
                    'sr': self.sr,
                    'y': y,
                }
            }
        }

        for corpus_id, corpus in enumerate(corpora):
            self.soundfiles[corpus_id] = {
                'source_root': corpus.source_root,
                'max_duration': corpus.max_duration,
                'sources': {}
            }
        CONSOLE.reset_bar('Finding matches for target segments:', max=len(target_analysis), item='segments')
        for x in target_analysis:
            matches = []
            for corpus_id, corpus in enumerate(corpora):
                nearest_neighbors = corpus.tree.knn(
                    x=x,
                    vector_path='features',
                    first_n=corpus.tree.leaf_size)
                for nn in nearest_neighbors:
                    # get id of source audio file
                    source_id = nn['value']['source']
                    self.soundfiles[corpus_id]['sources'][source_id] = corpus.soundfiles[source_id]
                    nn['value']['corpus'] = corpus_id
                    option = deepcopy(nn)
                    del option['value']['features']
                    matches.append(option)
            self.frames.append([x['value'] for x in sorted(matches, key=lambda x: x['cost'])])
            CONSOLE.bar.next()
        CONSOLE.bar.finish()

    def _serialize(self) -> dict:
        mosaic = deepcopy(vars(self))

        # if not portable, delete audio samples from soundfiles on write
        if not self.portable:
            for corpus_id in mosaic['soundfiles']:
                for source_id in mosaic['soundfiles'][corpus_id]['sources']:
                    del mosaic['soundfiles'][corpus_id]['sources'][source_id]['y']
        return mosaic

    def _summarize(self) -> dict:
        duration = str(datetime.timedelta(seconds=int(self.duration))) if self.duration else None
        num_corpora = 0
        num_sources = []
        for k in self.soundfiles:
            if k < 0:
                continue
            num_corpora += 1
            N = len(self.soundfiles[k]['sources'])
            info = f"{N} source(s) from corpus {num_corpora}"
            num_sources.append(info)

        return {
            "target": basename(str(self.target)),
            "portable": self.portable,
            "duration (H:M:S)": duration,
            "num. of corpora": num_corpora,
            "num. of sources": ", ".join(num_sources),
            "analysis features": ", ".join(self.features),
            "num. of grains": len(self.frames)
        }

    def _preload(self, obj: dict) -> dict:
        # reload soundfiles if non-portable
        if not obj['portable']:
            self.__load_soundfiles(obj['soundfiles'])
        return obj

    def read(self, file: str) -> Self:
        return super().read(file, warn_user=self.frames)

    def __load_soundfiles(self, soundfiles: Iterable) -> None:
        CONSOLE.counter.message = CONSOLE.log_subprocess('Loading audio files: ')
        for corpus_id in soundfiles:
            corpus = soundfiles[corpus_id]
            sources = corpus['sources']
            for source_id in sources:
                source = sources[source_id]
                if 'y' not in source:
                    path = join(corpus['source_root'], source['file'])
                    soundfiles[corpus_id]['sources'][source_id]['y'] = load(
                        path, sr=source['sr'], duration=corpus['max_duration'])[0]
                CONSOLE.counter.next()
        CONSOLE.counter.finish()

    @get_elapsed_time
    def to_audio(self,
                 # dynamic control parameters
                 fidelity: float | int | Envelope | Iterable = 1.0,
                 grain_dur: float | int | Envelope | Iterable = 0.1,
                 stretch_factor: float | int | Envelope | Iterable = 1.0,
                 onset_var: float | int | Envelope | Iterable = 0,
                 pan_depth: float | int | Envelope | Iterable = 5,
                 grain_env: str | Envelope | Iterable = "cosine",
                 corpus_weights: float | int | Envelope | Iterable = 1.0,

                 # static parameters
                 n_chans: int = 2,
                 sr: int | None = None,
                 win_length_res: int = 512) -> AudioBuffer:
        """
        Returns an *audio mosaic* as an ``AudioBuffer`` instance, based on several audio control parameters, such as `grain duration`, `onset variation`, `panning depth`, `grain envelope`, `grain duration`, `number of channels`, and more.

        fidelity: float | int | Envelope | Iterable = 1.0
            Normalized probablity (0.0 - 1.0) of choosing the best possible match from ``Corpus`` for each grain in ``target``.

        grain_dur: float | int | Envelope | Iterable = 0.1
            Grain duration in seconds.

        stretch_factor: float | int | Envelope | Iterable = 1.0
            Stretch factor of audio output (e.g., 1: original speed, 0.5: twice as fast, 2: twice as slow, etc.)

        onset_var: float | int | Envelope | Iterable = 0
            Grain onset variation in seconds.

        corpus_weights: float | int | Envelope | Iterable = 1.0,
            Probability of choosing grain from input ``Corpus`` (or ``list`` thereof) vs. from original audio ``target``.\n 
            Here are some examples that demonstrate the different ways in which this parameter can be specified:

                * ``int | float``: probability of choosing a grain from ``Corpus``. If more than one ``Corpus`` is given, the probability is equally distributed among them.
                * ``Envelope | List[float | int | tuple, ...]``: envelope of probabilities of choosing a grain from ``Corpus``. If a ``list`` of ``Corpus`` instances is given, the probability is equally distributed among them.
                * ``List[List[float | int | tuple, ...] | Envelope | (int | float), ...]``: ``list`` of envelopes of probabilities for chosing a grain from element at index ``i``, where ``i = 0`` is the ``target`` and the rest are all ``Corpus`` instances passed to ``corpus`` argument.

        pan_depth: float | int | Envelope | Iterable = 5
            Depth of contrast for channel panning

        grain_env: Envelope | str | Iterable = Envelope()
            Amplitude envelope for all audio grains.

        n_chans: int = 2
            Number of output audio channels.

        sr: int | None = None
            Sampling rate for audio output.

        win_length_res: int = 512
            Grain duration resolution in samples.

        """
        n_segments = len(self.frames)

        def as_points(param, N: int = n_segments) -> Points:
            """ resolve parameter into a ``Points`` instance based on type """
            return object_to_points(param, N)

        def parse_corpus_weights_param(param: Envelope | float | int) -> np.ndarray:
            """ Generates corpus weights table based on param type """
            even_weights = True
            num_corpora = len(self.soundfiles.keys()) - 1
            # when weigths are meant to control target vs all corpora
            # NOTE: tuple instance used here for backwards typing compatibility, needed for non-native classes
            if isinstance(param, (Envelope, float, int)) or (
                    isinstance(param, Iterable) and all([isinstance(p, int | float | tuple) for p in param])):
                wet_mix = as_points(param).clip(min=0.0, max=1.0)
                mix_table = np.array([1.0-wet_mix] + [wet_mix for _ in range(num_corpora)]).T
            # when weights are meant to control target and corpora individually
            else:
                if len(param) != num_corpora + 1:
                    CONSOLE.error(
                        ValueError,
                        f'The number of items in corpus_weights argument must be {num_corpora + 1}, where the first element is the target')
                mix_table = np.array([as_points(env).clip(min=0.0, max=1.0) for env in param]).T
                even_weights = False
            return mix_table / mix_table.sum(axis=1)[:, np.newaxis], even_weights

        def preprocess_samples(n_chans: int, sr: int) -> None:
            """ resamples files with conflicting sampling rates """
            soundfiles = deepcopy(self.soundfiles)
            CONSOLE.reset_counter('Preprocessing audio files: ')
            for corpus_id in soundfiles:
                sources = soundfiles[corpus_id]['sources']
                for source_id in sources:
                    source = sources[source_id]
                    y = source['y']
                    sr_ratio = sr/source['sr']
                    if source['sr'] != sr:
                        y = resample_array(y, int(len(y) * sr_ratio))
                    soundfiles[corpus_id]['sources'][source_id]['y'] = np.repeat(np.array([y]).T, n_chans, axis=1)
                    CONSOLE.counter.next()
            CONSOLE.counter.finish()
            return soundfiles

        CONSOLE.log_process(f'\N{brain} Generating audio from mosaic target: {basename(self.target)}...').print()
        # playback ratio
        sr, sr_ratio = (self.sr, 1) if not sr else (sr, sr/self.sr)
        hop_length = int(self.hop_length * sr_ratio)

        soundfiles = preprocess_samples(n_chans=n_chans, sr=sr)

        # DYNAMIC CONTROL TABLES
        CONSOLE.log_subprocess('Creating parameter envelopes...').print()

        corpus_weights_table, even_weights = parse_corpus_weights_param(corpus_weights)

        win_length_table = (as_points(grain_dur) * sr).quantize(win_length_res).astype('int64')

        max_win_length = np.amax(win_length_table) + win_length_res
        win_lengths = np.arange(win_length_res, max_win_length, win_length_res, dtype='int64')

        samp_onset_table = (as_points(stretch_factor)
                            * hop_length).quantize().concat([0], prepend=True).astype('int64').cumsum()[:-1]

        # apply onset variation to samp_onset_table
        samp_onset_var_table = (np.random.rand(n_segments) - 0.5) * as_points(onset_var) * (sr // 2)
        samp_onset_table += samp_onset_var_table.astype('int64')
        samp_onset_table[samp_onset_table < 0] = 0

        # compute amplitude windows
        windows = [as_points(grain_env, wl).wrap().T.replicate(n_chans, axis=1) for wl in win_lengths]

        # compute panning table
        pan_depth_table = as_points(pan_depth).wrap().T.replicate(n_chans, axis=1)
        pan_table = Points(np.linspace(0, 1, n_chans)).wrap().replicate(n_segments, axis=0) - np.random.rand(n_segments, 1)
        pan_table = 1 / (2**(pan_depth_table * pan_table.abs()))
        pan_table /= pan_table.sum(axis=1)[:, np.newaxis]

        # make buffer array
        buffer = np.empty(shape=(int(np.amax(samp_onset_table) + np.amax(win_length_table)), n_chans))
        buffer.fill(0)

        CONSOLE.reset_bar('Concatenating grains:', max=len(self.frames), item='grains')

        fidelity_table = as_points(fidelity).clip(0.0, 1.0)

        for n, (frames, grain_onset, win_length, pan_value, corpora_weights, fidelity_value) in enumerate(
                zip(self.frames, samp_onset_table, win_length_table, pan_table, corpus_weights_table, fidelity_table)):
            target_weight = corpora_weights[0]
            if random() > target_weight:
                # use choices from a single corpus if corpora_mix was provided
                amp = 1.0
                candidates = frames
                if not even_weights:
                    corpus_id = choices([x for x in self.soundfiles.keys() if x != -1], weights=corpora_weights[1:])[0]
                    candidates = [x for x in candidates if x['corpus'] == corpus_id]

                num_candidates = max(1, int(len(candidates) * (1 - fidelity_value)))
                weights = np.linspace(1.0, 0.0, num_candidates)
                f = choices(candidates[:num_candidates], weights=weights)[0]
            else:
                f = {
                    'corpus': -1,
                    'source': 0,
                    'marker': n * self.hop_length,
                }
                amp = target_weight
            source_id = f['source']
            corpus_id = f['corpus']
            source = soundfiles[corpus_id]['sources'][source_id]['y']
            source_sr_ratio = sr/soundfiles[corpus_id]['sources'][source_id]['sr']
            max_idx = len(source) - 1
            grain_start = int(f['marker'] * source_sr_ratio)
            grain_end = min(max_idx, grain_start+win_length)
            grain_size = round((grain_end-grain_start) / win_length_res) * win_length_res
            grain_end = grain_start+grain_size
            if grain_size > 0 and grain_end <= max_idx:
                idx = int(np.where(win_lengths == grain_size)[0])
                window = windows[idx]
                grain = source[grain_start:grain_end] * window * pan_value * amp
                buffer[grain_onset:grain_onset+grain_size] = buffer[grain_onset:grain_onset+grain_size] + grain
            CONSOLE.bar.next()

        CONSOLE.bar.finish()

        # return normalized buffer
        return AudioBuffer(y=(buffer / np.amax(np.abs(buffer))) * np.sqrt(0.5), sr=sr)



# typing
from __future__ import annotations

# ui
from kivy.uix.button import Button
from kivy.uix.togglebutton import ToggleButton
from kivy.properties import ListProperty

import os
from kivy.core.window import Window


class BaseButton(Button):
    bg_color = ListProperty((0, 0, 0, 0))
    bg_color_down = ListProperty((0, 0, 0, 0))

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        Window.bind(mouse_pos=self.on_hover)

    def on_hover(self, w, p):
        """ Highlighted button color on hover """
        if self.disabled:
            return
        if not hasattr(self, 'bg_color_hover'):
            self.bg_color_normal = self.bg_color
            self.bg_color_hover = tuple(min(1, x*1.25) for x in self.bg_color)
        self.bg_color = self.bg_color_hover if self.collide_point(*p) else self.bg_color_normal


class LargeButton(BaseButton):
    pass


class LargeDangerButton(LargeButton):
    pass


class LargeSuccessButton(LargeButton):
    pass


class MenuItem(ToggleButton):
    def __init__(self, value: str = 'item', **kwargs) -> None:
        self.value = value
        super().__init__(**kwargs)
        self.text = os.path.basename(self.value)



# typing
from __future__ import annotations

# import all widgets
from .__gui_init__ import *
from .buttons import *
from .corpus import *
from .audio import *
from .mosaic import *

# gamut
from ..__version__ import __version__
from .theme import Theme
from .utils import log_message

# kivy imports
from kivy.properties import ObjectProperty
from kivy.uix.widget import Widget
from kivy.app import App
from kivy.clock import Clock


class Main(Widget):
    """
    Main UI Widget, containing all modules.
    Child widgets are included through the .kv file
    """
    console = ObjectProperty(None)
    mosaic_module = ObjectProperty(None)
    corpus_module = ObjectProperty(None)

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        Clock.schedule_once(lambda _: log_message("GAMuT session intialized"), 1)


class GUI(App):
    """
    GAMuT Graphical User Interface class
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.theme = Theme()
        self.version = f"v{__version__}"

    def build(self) -> Widget:
        """ Returns root widget """
        self.title = 'GAMuT user interface'
        self.icon = 'data/images/icon.png'
        return Main()

    def run(self, test: int | None = None) -> None:
        if test:
            Clock.schedule_once(lambda _: self.stop(), test)
        super().run()



from __future__ import annotations
import numpy as np
from .utils import get_nested_value, set_nested_value
from .config import CONSOLE
from random import randint
from collections.abc import Iterable


class KDTree:
    """
    k-dimensional binary search tree

    Attributes
    ----------
    leaf_size: int = 10
        Maximum number of data items per leaf.
    """

    def __init__(self, leaf_size: int = 10) -> None:
        self.leaf_size = leaf_size
        self.k = None
        self.min = None
        self.min = None
        self.norm = None
        self.data = None

    def __update_minmax(self, vector: np.ndarray) -> None:
        self.min = np.min(np.array([self.min, vector]), axis=0)
        self.max = np.max(np.array([self.max, vector]), axis=0)

    def __build(self, data: list, vector_path: str, k: int = 0) -> dict:
        data_size = len(data)
        if (data_size <= self.leaf_size):
            CONSOLE.bar.next(data_size)
            for d in data:
                self.__update_minmax(get_nested_value(d, vector_path))
            return {
                'leaf': data,
            }
        data.sort(key=lambda x: get_nested_value(x, vector_path)[k])
        middle_index = len(data) // 2
        next_k = (k + 1) % self.k
        vector = get_nested_value(data[middle_index], vector_path)
        return {
            'node': {
                'k': k,
                'value': vector[k]
            },
            0: self.__build(data[:middle_index], vector_path, next_k),
            1: self.__build(data[middle_index:], vector_path, next_k),
        }

    def build(self, data: list | np.ndarray, vector_path: str) -> None:
        # get tree dimensions
        self.k = len(get_nested_value(data[0], vector_path))

        # initialize min, max, and norm arrays
        self.min = np.zeros(self.k)
        self.min.fill(np.inf)

        self.max = np.zeros(self.k)
        self.max.fill(-np.inf)

        self.norm = np.ones(self.k)

        CONSOLE.reset_bar('Classifying audio grains:', max=len(data), item='grains')
        # build binary tree and fit data
        self.data = self.__build(data, vector_path, k=randint(0, self.k-1))
        CONSOLE.bar.finish()
        self.norm = self.max - self.min
        self.__fit(vector_path)

    def __normalize_input(self, x: np.ndarray) -> np.ndarray:
        return (x - self.min) / self.norm

    def __fit(self, vector_path: str) -> None:
        """ recursively normalize tree data """
        def fit(tree, vector_path):
            if 'node' in tree:
                k = tree['node']['k']
                tree['node']['value'] = (
                    tree['node']['value'] - self.min[k]) / self.norm[k]
                fit(tree[0], vector_path)
                fit(tree[1], vector_path)
                return
            for leaf_item in tree['leaf']:
                vector = get_nested_value(leaf_item, vector_path)
                set_nested_value(leaf_item, vector_path,
                                 self.__normalize_input(vector))
        fit(self.data, vector_path)

    def __euclidean_distance(self, a: np.ndarray, b: np.ndarray) -> np.ndarray:
        return np.sqrt(np.sum((a-b)**2))

    def knn(self, x: np.ndarray, vector_path: str, first_n: int = 10) -> Iterable:
        data = self.data
        data_point = self.__normalize_input(x)
        while 'node' in data:
            node = data['node']
            data = data[int(data_point[node['k']] >= node['value'])]
        return sorted(
            [
                {
                    'cost': self.__euclidean_distance(
                        data_point,
                        get_nested_value(leaf_item, vector_path)),
                    'value': leaf_item
                } for leaf_item in data['leaf']
            ], key=lambda x: x['cost'])[:first_n]

    def read(self, tree: dict) -> None:
        for attr in tree:
            hasattr(self, attr) and setattr(self, attr, tree[attr])



# typing
from __future__ import annotations
from collections.abc import Iterable, Callable

# misc
from ast import literal_eval
import re

# kivy
from kivy.factory import Factory
from kivy.app import App
from kivy.uix.popup import Popup


class ConsoleLog(Factory.TextLabel):
    def __init__(self, *args, index, **kwargs) -> None:
        self.index = index
        super().__init__(*args, **kwargs)


class UserConfirmation(Popup):
    def __init__(self, on_confirm, long_text: str | None, **kwargs) -> None:
        self.on_confirm = on_confirm
        self.long_text = '{}\nDo you want to proceed?'.format(long_text or "You\'re about to delete these item(s).")
        super().__init__(**kwargs)


APP = None


def get_log_style(key) -> dict:
    app = get_app()
    log_styles = {
        'normal': {
            'color': app.theme.colors.txt3,
        },
        'error': {
            'color': app.theme.colors.danger,
            'bold': True
        },
        'success': {
            'color': app.theme.colors.txt_hl
        },
    }
    return log_styles[key]


def get_app() -> App:
    global APP
    if not APP:
        APP = App.get_running_app()
    return APP


def log_message(text: str | Iterable, log_type: str = 'normal') -> None:
    """ Logs a message in the GUI console window """
    root = get_app().root
    logs = [text] if isinstance(text, str) else text
    style = get_log_style(log_type)
    for log in logs:
        l = ConsoleLog(text=log, index=len(root.console.children), **style)
        root.console.add_widget(l)
    if len(root.console.children) > 16:
        root.console.parent.scroll_to(l)


def capture_exceptions(func) -> Callable:
    """ Decorator to prevent crashing by printing error in GUI console window """
    def decorator(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            error_type = type(e).__name__
            msg = remove_ansi(str(e))
            log_message([f"{error_type}: {msg}"], 'error')
    return decorator


def log_done(func) -> Callable:
    """ Decorator for logging successful callback execution """
    def decorator(*args, **kwargs):
        out = func(*args, **kwargs)
        log_message("Done!", 'success')
        return out
    return decorator


def remove_ansi(text) -> str:
    """ Removes ANSI stuff from string """
    ansi_pattern = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    return ansi_pattern.sub('', text)


def parse_param_string(raw_string: str) -> Iterable:
    """ Parses a lisp-like formatted list string as Python list """
    # remove spacing at both ends
    value = raw_string.strip()

    # return if a it's simple string
    if value.isalpha():
        return value

    # remove parentheses-adjacent spaces between
    paren_spaces = re.compile(r' +(?=\))|(?<=\() +')
    value = paren_spaces.sub('', value)

    # replace duplicated spaces with single ones
    dup_spaces = re.compile(r' +')
    value = dup_spaces.sub(' ', value)

    # replace chars before literal eval
    for before, after in [('(', '['), (' ', ','), (')', ']')]:
        value = value.replace(before, after)
    return literal_eval(value)



from __future__ import annotations
from time import time
from .sys import Console
from typing import Any, Callable
import os

FILE_EXT = '.gamut'
AUDIO_FORMATS = ['.wav', '.aif', '.aiff', '.mp3']
MIME_TYPES = ['audio/x-wav', 'audio/x-aiff', 'audio/mpeg']
AUDIO_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'gui/data/audio/')
CONSOLE = Console()
ANALYSIS_TYPES = ['timbre', 'pitch']
ENVELOPE_TYPES = [
    'barthann',
    'bartlett',
    'blackman',
    'blackmanharris',
    'bohman',
    'boxcar',
    'chebwin',
    'cosine',
    'dpss',
    'exponential',
    'flattop',
    'gaussian',
    'general_cosine',
    'general_gaussian'
    'general_hamming'
    'hamming'
    'hann'
    'kaiser_bessel_derived'
    'lanczos'
    'nuttall',
    'parzen',
    'taylor',
    'triang',
    'tukey',
]


def get_elapsed_time(func: Callable) -> Callable:
    """ Decorator to log how long a given function takes to run """
    def decorator(*args, **kwargs) -> Any:
        st = time()
        output = func(*args, **kwargs)
        CONSOLE.elapsed_time(st).print()
        return output
    return decorator



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
import codecs


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


def get_logo():
    here = os.path.dirname(__file__)
    return os.path.join(os.path.realpath(here), '../../gamut/gui/data/images/logo.png')


sys.path.insert(0, os.path.abspath('../..'))

project = 'GAMuT'
copyright = '2023, Felipe Tovar-Henao'
author = 'Felipe Tovar-Henao'
version = "v{}".format(get_version('../../gamut/__version__.py'))

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc']

templates_path = ['_templates']
exclude_patterns = []
extensions = [
    'sphinx_copybutton',
    'sphinx.ext.autodoc',
]
html_favicon = '_static/favicon.ico'


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']
html_logo = get_logo()
html_theme_options = {
    'logo_only': True,
    'display_version': True,
}
autodoc_typehints = "none"

pygments_style = 'monokai'



# typing
from __future__ import annotations

# ui
from kivy.config import Config
from kivy.resources import resource_add_path
from kivy.lang import Builder

# gamut
from ..sys import set_vebosity

# misc.
import os

# disable GAMuT CLI messages
set_vebosity(False)

# set default kivy window configuration on open
for key, val in [('resizable', False), ('width', '1360'), ('height', '850'), ('top', '100'), ('left', '100')]:
    Config.set('graphics', key, val)


def __load_kv_files__() -> None:
    """ Points Kivy to .kv file directory """
    root_dir = os.path.dirname(__file__)
    kv_dir = os.path.join(root_dir, 'data/widgets')
    resource_add_path(root_dir)

    for file in os.listdir(kv_dir):
        Builder.load_file(os.path.join(kv_dir, file))


# load all available .kv files
__load_kv_files__()



import os.path
import codecs
from setuptools import setup, find_namespace_packages
from pathlib import Path

long_description = (Path(__file__).parent / "README.md").read_text()

MODULE = 'gamut'
DESCRIPTION = 'Granular Audio Musaicing Toolkit for Python'


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


def get_install_requires():
    excluded = ['sphinx']
    with open('requirements.txt', 'r') as f:
        return [line for line in f.read().splitlines() if all(x not in line.lower() for x in excluded)]


setup(
    name=MODULE,
    version=get_version('gamut/__version__.py'),
    author='Felipe Tovar-Henao',
    author_email='<felipe.tovar.henao@gmail.com>',
    description=DESCRIPTION,
    url='https://felipetovarhenao.github.io/gamut',
    packages=find_namespace_packages(),
    license='OSI Approved :: ISC License (ISCL)',
    include_package_data=True,
    entry_points={
        'console_scripts': ['gamut=gamut:cli', 'gamut-ui=gamut:gui']
    },
    install_requires=get_install_requires(),
    keywords=['DSP', 'audio musaicking', 'granulation', 'machine learning',
              'ML', "MIR", 'music', 'sound design', 'concatenative synthesis'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Other Audience',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python'
    ],
    long_description=long_description,
    long_description_content_type='text/markdown'
)



# typing
from collections.abc import Iterable

# gamut
from .utils import catch_keyboard_interrupt


def print_success(text) -> None:
    """ Prints green-colored message """
    print(f"\N{check mark} \033[32;1m{text}\033[0m")


def print_warning(text) -> None:
    """ Prints yellow-colored message """
    print(f"\N{warning sign} \033[33;1m{text}\033[0m")


def print_error(error) -> None:
    """ Prints red-colored message """
    print(f"\N{skull} \033[31;1m{error}\033[0m")
    exit()


def is_kivy_installed() -> bool:
    """ Check if kivy is installed """
    import pkg_resources
    return 'kivy' in list(map(lambda x: x.key, pkg_resources.working_set))


def gui() -> None:
    """ Entry points to launch GUI """

    # Check if kivy has been installed
    if not is_kivy_installed():
        import subprocess
        import threading
        from sys import executable

        # try installing on separate thread and wait until done
        print_warning("Attempting to install missing dependencies...")
        t = threading.Thread(target=lambda: subprocess.run([executable, '-m', 'pip', 'install', 'kivy[base]']))
        t.start()
        t.join()

        # run with --retry flag to prevent infinite recursion (see conditional in cli())
        print_success("Running again...")
        subprocess.run(['gamut', '--gui', '--retry'])
        exit()

    import os
    os.environ["KIVY_NO_CONSOLELOG"] = "1"

    print_success('Launching user interface...')
    from .gui import GUI
    GUI().run()


@catch_keyboard_interrupt()
def cli():
    """ CLI entry point """
    from argparse import ArgumentParser
    from os import chdir, makedirs, remove, mkdir, getcwd
    from os.path import realpath, join, exists, splitext, basename
    from shutil import rmtree, copyfileobj
    from subprocess import run
    from pathlib import Path
    import json
    import requests

    # ------------------------------------- #
    # HELPER FUNCTIONS
    # ------------------------------------- #

    def parse_params(raw_params: Iterable) -> dict:
        params = {}
        if not raw_params:
            return params
        for param in raw_params:
            key, value = param.split("=")
            try:
                value = float(value)
            except:
                pass
            if key in params:
                if not isinstance(params[key], list):
                    params[key] = [params[key]]
                params[key].append(value)
                continue
            params[key] = value
        return params

    def safe_chdir(dest: str) -> None:
        if not exists(dest):
            return
        chdir(dest)

    def abs_path(path: str, ext: str) -> str:
        return realpath(splitext(path)[0] + ext)

    def write_file(path: str, obj: object, ext: str, script_block: str) -> None:
        out = abs_path(path, ext)
        if not NO_CACHE:
            CACHED_OBJECTS[out] = obj
        if script_block not in SKIP_WRITE:
            if exists(out):
                remove(out)
            obj.write(out)

    def clean_path(file: str) -> str:
        path = realpath(file)
        if not exists(path):
            print_error(f"{path} does not exist.")
        return path

    def check_subdirectories(script: str) -> None:
        subdirs = [AUDIO_DIR, CORPUS_DIR, MOSAIC_DIR]
        missing = []
        for subdir in subdirs:
            if not exists(subdir):
                missing.append(basename(subdir))
        num_missing = len(missing)
        if num_missing == 0:
            return
        if num_missing > 0:
            for p in [Path(script).parent.parent.absolute(), Path(script).parent.absolute()]:
                found = True
                for subdir in subdirs:
                    if not exists(join(p, basename(subdir))):
                        found = False
                        break
                if found:
                    define_directories(p)
                    return
        if num_missing == len(subdirs):
            print_warning('You seem to be running this script from outside a workspace folder, which might result in unexpected behaviors.')
        elif num_missing > 0:
            print_warning(
                f'Your current directory is missing the following workspace folders:\n\t{", ".join(missing)}\nThis might result in unexpected behaviors.')
        answer = input("Would you like to continue (y/n)?")
        if answer.lower() in ['n', 'no']:
            exit()

    def define_directories(root: str) -> None:
        global ROOT_DIR
        global MOSAIC_DIR
        global CORPUS_DIR
        global AUDIO_DIR
        global SCRIPTS_DIR
        global TEST_NAME
        global TEST_SCRIPT_DIR
        ROOT_DIR = realpath(root)
        MOSAIC_DIR = join(ROOT_DIR, 'mosaics')
        CORPUS_DIR = join(ROOT_DIR, 'corpora')
        AUDIO_DIR = join(ROOT_DIR, 'audio')
        SCRIPTS_DIR = join(ROOT_DIR, 'scripts')
        TEST_NAME = 'test'
        TEST_SCRIPT_DIR = join(SCRIPTS_DIR, f'{TEST_NAME}.json')

    def create_new_template(template: str) -> None:
        chdir(ROOT_DIR)
        out_path = splitext(template)[0]
        name = basename(out_path)
        script = {
            "corpus": {
                "name": f"{name}-corpus",
                "source": [
                    join(AUDIO_DIR, DEFAULT_SOURCE)
                ],
                "features": [
                    "timbre"
                ]
            },
            "mosaic": {
                "name": f"{name}-mosaic",
                "target": join(AUDIO_DIR, DEFAULT_TARGET),
                "corpus": [
                    f"{name}-corpus"
                ]
            },
            "audio": {
                "name": f"{name}-audio",
                "mosaic": f"{name}-mosaic",
                "fidelity": 1.0,
                "grain_dur": 0.1,
                "grain_env": "cosine",
                "corpus_weights": 1.0,
                "stretch_factor": 1.0,
                "pan_depth": 3,
                "onset_var": 0.0,
                "n_chans": 2,
                "sr": 44100,
            }
        }
        with open(out_path + '.json', 'w') as f:
            json.dump(script, f, indent=4)

    # ------------------------------------- #
    # DEFINE FOLDER STRUCTURE
    # ------------------------------------- #
    define_directories(getcwd())

    # ------------------------------------- #
    # PARSE CLI ARGUMENTS
    # ------------------------------------- #
    CACHED_OBJECTS = {}
    SCRIPT_MODES = ['corpus', 'mosaic', 'audio']
    DEFAULT_SOURCE = 'source.mp3'
    DEFAULT_TARGET = 'target.mp3'

    parser = ArgumentParser(prog='GAMuT parser',
                            description="Command-line utility for creating GAMuT audio musaicings with JSON files",
                            epilog='To learn more, visit https://felipe-tovar-henao.com/gamut')
    parser.add_argument('-v', '--version',
                        action='store_true',
                        help='get current version of GAMuT')
    parser.add_argument('--test',
                        action='store_true',
                        help='runs GAMuT package test.')
    parser.add_argument('--gui',
                        action='store_true',
                        help='open GAMuT graphical user interface.')
    parser.add_argument('--retry',
                        action='store_true',
                        help='Flag to signal attempt to re-run UI after kivy installation attempt.')
    parser.add_argument('--no-check',
                        action='store_true',
                        help='Disable directory checking when reading a script')
    parser.add_argument('--no-cache',
                        action='store_true',
                        help='Disable pipeline caching')
    parser.add_argument('--no-download',
                        action='store_true',
                        help='Disable audio when using --init')
    parser.add_argument('-i', '--init',
                        nargs='?',
                        const=ROOT_DIR,
                        help='initializes a project folder at the specified directory',
                        type=str)
    parser.add_argument('-s', '--script',
                        help="set JSON file to use as input settings for GAMuT",
                        type=str)
    parser.add_argument('--summarize',
                        help="show summary of a .gamut file",
                        type=str)
    parser.add_argument('-p', '--play',
                        action='store_true',
                        help="enable audio playback after script runs")
    parser.add_argument('--no-verbose',
                        action='store_true',
                        help="disables console printing")
    parser.add_argument('--skip',
                        nargs='+',
                        help="skip one or more blocks from the script")
    parser.add_argument('--skip-write',
                        nargs='+',
                        help="skip writing to disk one or more blocks from the script")
    parser.add_argument('-t', '--template',
                        nargs='?',
                        const=TEST_SCRIPT_DIR,
                        help="generate a JSON script template",
                        type=str)
    parser.add_argument('--source',
                        nargs='+',
                        help="path to corpus source(s)")
    parser.add_argument('--params',
                        nargs='+',
                        help="audio control parameters")
    parser.add_argument('--features',
                        nargs='+',
                        help="audio features to base audio musaicking on",
                        choices=['pitch', 'timbre'])
    parser.add_argument('--target',
                        help="path to audio target",
                        type=str)
    parser.add_argument('--audio',
                        default="gamut-output.wav",
                        help='path to audio output',
                        type=str)

    args = parser.parse_args()

    NO_CACHE = args.no_cache
    SKIP_WRITE = args.skip_write or []
    raw_musaicing_args = [args.source, args.target]

    if len(SKIP_WRITE) > 0 and NO_CACHE:
        print_error("You can't use --skip-write and --no-cache at the same time.")

    if args.no_verbose:
        from .sys import set_vebosity
        set_vebosity(False)

    # ------------------------------------- #
    # PRINT GAMUT PACKAGE VERSION
    # ------------------------------------- #
    if args.version:
        from .__version__ import __version__
        print_success(f'GAMuT v{__version__}')

    # ------------------------------------- #
    # OPEN GRAPHICAL USER INTERFACE
    # ------------------------------------- #
    elif args.gui:
        if not is_kivy_installed() and args.retry:
            print('Unable to install kivy library. Please visit https://kivy.org/doc/stable/gettingstarted/installation.html#install-pip to install it manually and try again afterwards.')
        import subprocess
        subprocess.run(['gamut-ui'])

    # ------------------------------------- #
    # RUN PACKAGE TEST
    # ------------------------------------- #
    elif args.test:
        TEST_DIR = realpath('.gamut-test')
        if exists(TEST_DIR):
            rmtree(TEST_DIR)
        mkdir(TEST_DIR)
        safe_chdir(TEST_DIR)
        print_success("Running test...")
        run(['gamut', '--init', '--no-verbose'])
        run(['gamut', '--script', join(basename(SCRIPTS_DIR), f"{TEST_NAME}.json"), '--no-cache', '--no-verbose'])
        print_success("Done")
        rmtree(TEST_DIR)

    # ------------------------------------- #
    # PROCESS RAW INPUT
    # ------------------------------------- #
    elif any(raw_musaicing_args):
        if not all(raw_musaicing_args):
            print_error("You must provide both source and target")
        params = parse_params(args.params)
        source, target = raw_musaicing_args
        target = clean_path(target)
        corpus_params = {
            'source': [clean_path(s) for s in source]
        }
        if args.features:
            corpus_params['features'] = args.features

        impulse_response = params.pop('impulse_response', None)
        if impulse_response:
            convolve_params = {
                'impulse_response': clean_path(impulse_response)
            }
            convolve_mix = params.pop('convolve_mix', None)
            if convolve_mix:
                convolve_params['mix'] = convolve_mix

        from .features import Corpus, Mosaic

        corpus = Corpus(**corpus_params)
        mosaic = Mosaic(target=target, corpus=corpus)
        audio = mosaic.to_audio(**params)
        write_file(args.audio, audio, '.wav', 'audio')
        if impulse_response:
            audio.convolve(**convolve_params)
        if args.play:
            audio.play()

    # ------------------------------------- #
    # INITIALIZE WORKSPACE
    # ------------------------------------- #
    elif args.init:
        if realpath(args.init) != ROOT_DIR:
            define_directories(args.init)
        for x in [MOSAIC_DIR, CORPUS_DIR, AUDIO_DIR, SCRIPTS_DIR]:
            if not exists(x):
                makedirs(x)
        create_new_template(TEST_SCRIPT_DIR)

        base_url = 'https://d2cqospqxtt8fw.cloudfront.net/personal-website/gamut/'
        filenames = [DEFAULT_SOURCE, DEFAULT_TARGET]

        if not args.no_download:
            try:
                for f in filenames:
                    file_path = join(AUDIO_DIR, f)
                    if exists(file_path):
                        continue
                    with requests.get(base_url + f, stream=True) as r:
                        with open(file_path, 'wb') as f:
                            copyfileobj(r.raw, f)
            except:
                print('\tWarning: Unable to download audio examples')
        if not args.no_verbose:
            print_success(f"Your GAMuT project folder is ready! Try running:\n\tgamut --script scripts/{TEST_NAME}.json --play")

    # ------------------------------------- #
    # CREATE TEMPLATE
    # ------------------------------------- #

    elif args.template:
        safe_chdir(SCRIPTS_DIR)
        create_new_template(args.template)

    # ------------------------------------- #
    # SUMMARIZE GAMUT FILE
    # ------------------------------------- #

    elif args.summarize:
        gamut_file = args.summarize
        if not exists(gamut_file):
            print_error(f'{gamut_file} does not exist.')

        if splitext(gamut_file)[1] != '.gamut':
            print_error(f'{gamut_file} is not a .gamut file')

        from .features import Corpus, Mosaic

        for obj in [Corpus(), Mosaic()]:
            try:
                obj.read(gamut_file).summarize()
            except:
                continue
            break

    # ------------------------------------- #
    # PROCESS SCRIPT
    # ------------------------------------- #

    elif args.script:
        script_path = clean_path(args.script)

        if not args.no_check:
            check_subdirectories(script_path)

        with open(file=script_path, mode='r',) as f:
            script = json.loads(f.read())

        # validate skipped blocks and warn if they don't exist
        skipped_blocks = args.skip if args.skip else []
        for skip_list in [skipped_blocks, SKIP_WRITE]:
            for skipped in skip_list:
                if skipped not in script:
                    print_warning(f"trying to skip a block that isn't included in the script: \"{skipped}\". ignoring...")

        # create mapping of script blocks to their types
        block_type_map = {}
        for script_block in script:
            fail = True
            for sm in SCRIPT_MODES:
                if script_block.startswith(sm):
                    fail = False
                    if script_block not in skipped_blocks:
                        block_type_map[script_block] = sm
                    break
            if fail:
                print_error(
                    f'"{script_block}" is not a valid GAMuT script block. It should start with one of the following: {SCRIPT_MODES}')

        from .features import Corpus, Mosaic

        # process each block
        for script_block in block_type_map:
            params = script[script_block]
            block_type = block_type_map[script_block]

            name = params.pop('name', None)
            output_name = name if name else splitext(basename(script_path))[0] + f'-{script_block}'

            if block_type == 'corpus':
                safe_chdir(AUDIO_DIR)
                for i, source in enumerate(params['source']):
                    params['source'][i] = clean_path(source)
                safe_chdir(ROOT_DIR)
                c = Corpus(**params)
                safe_chdir(CORPUS_DIR)
                write_file(output_name, c, '.gamut', script_block)

            elif block_type == 'mosaic':
                for x in ['corpus', 'target']:
                    if x not in params:
                        print_error(f'You forgot to specify a {x} in your {block_type} script.')

                # clean target path
                safe_chdir(AUDIO_DIR)
                params['target'] = clean_path(params['target'])

                # clean beat unit
                if 'beat_unit' in params:
                    try:
                        if "/" in params['beat_unit']:
                            a, b = params['beat_unit'].split('/')
                            params['beat_unit'] = int(a) / int(b)
                        else:
                            params['beat_unit'] = int(params['beat_unit'])
                    except:
                        print_error('Invalid "beat_unit" value')

                # clean corpus paths
                safe_chdir(CORPUS_DIR)
                corpus_paths = params.pop('corpus')
                corpora = []
                for c in corpus_paths:
                    cpath = abs_path(c, '.gamut')
                    if cpath in CACHED_OBJECTS:
                        corpora.append(CACHED_OBJECTS[cpath])
                        continue
                    cpath = clean_path(cpath)
                    corpora.append(Corpus().read(cpath))

                # build mosaic
                safe_chdir(ROOT_DIR)
                m = Mosaic(corpus=corpora, **params)

                # write to disk
                safe_chdir(MOSAIC_DIR)
                write_file(output_name, m, '.gamut', script_block)

            elif block_type == 'audio':
                mosaic_file = params.pop('mosaic')
                convolve = params.pop('convolve', None)

                # clean mosaic path
                safe_chdir(MOSAIC_DIR)
                mosaic_path = abs_path(mosaic_file, '.gamut')
                m = Mosaic().read(clean_path(mosaic_path)
                                  ) if mosaic_path not in CACHED_OBJECTS else CACHED_OBJECTS[mosaic_path]

                # create audio mosaic
                safe_chdir(ROOT_DIR)
                audio = m.to_audio(**params)
                safe_chdir(AUDIO_DIR)

                # clean convolution
                if convolve:
                    if 'impulse_response' not in convolve:
                        print_error('To apply audio convolution, you must provide an inpulse response')
                    convolve['impulse_response'] = clean_path(convolve['impulse_response'])
                    audio.convolve(**convolve)
                write_file(output_name, audio, '.wav', script_block)
                if args.play:
                    audio.play()
    else:
        print_error("You didn't provide any required arguments. Use -h or --help to learn more.")



# typing
from __future__ import annotations

# ui
from kivy.metrics import dp


def rgba(hex: str) -> tuple:
    """ HEX to RGBA conversion """
    return *tuple(int(hex[int(hex[0] == '#'):][i:i+2], 16)/255.0 for i in (0, 2, 4)), 1


class ColorPalette:
    """ Entrypoint for all RGBA-formatted theme colors """

    def __init__(self) -> None:
        self.bg1 = rgba("1d1f25")
        self.bg2 = rgba("272931")
        self.bg3 = rgba("30323b")
        self.bg4 = rgba("3E404A")
        self.txt1 = rgba("e1e1e1")
        self.txt2 = rgba("d6d6d6")
        self.txt3 = rgba("c1c1c1")
        self.txt_hl = rgba("D8E4FF")
        self.txt_neg = rgba("ffffff")
        self.primary = rgba("4d5360")
        self.secondary = rgba("5b6476")
        self.success = rgba("60799f")
        self.success_hl = rgba("718cb5")
        self.danger = rgba("A05F5F")
        self.danger_hl = rgba("E97575")
        self.border = rgba("474A53")
        self.disabled_mask = (0, 0, 0, 0.4)


class Font:
    """
    Utility class to centralize font sizes for GUI
    """

    def __init__(self,
                 font_name: str = "Roboto",
                 xs: int | float = 9,
                 sm: int | float = 10,
                 ms: int | float = 12,
                 md: int | float = 14,
                 ml: int | float = 16,
                 lg: int | float = 18,
                 xl: int | float = 20,
                 xxl: int | float = 22) -> None:
        self.name = font_name
        self._xs = dp(xs)
        self._sm = dp(sm)
        self._ms = dp(ms)
        self._md = dp(md)
        self._ml = dp(ml)
        self._lg = dp(lg)
        self._xl = dp(xl)
        self._xxl = dp(xxl)

    def size(self, size: str | None = None) -> float:
        """ Font size getter """
        return getattr(self, f"_{size or 'md'}")


class Theme:
    """ Entrypoint for GUI theme properties """

    def __init__(self) -> None:
        self.colors = ColorPalette()
        self.font = Font()
        self.spacing = dp(10)
        self.line_width = dp(1)

    def pad(self, n: int = 1) -> float:
        return dp(3*n)



from __future__ import annotations
import numpy as np
from time import time
from .utils import resample_array
from progress import bar, counter, Infinite
from typing_extensions import Any


def set_vebosity(verbose: bool) -> None:
    """ Set verbosity of Console class"""
    Console.set_verbose(verbose)


class Console:
    """
    Utility class for logging ANSI-colored text in the console.
    """

    VERBOSE = True
    CALLED = False

    class Log(str):
        """ Dummy subclass of str to include print as method """
        VERBOSE = True

        @classmethod
        def set_verbose(cls, verbose):
            cls.VERBOSE = verbose

        def print(self) -> None:
            if not self.VERBOSE:
                return
            print(self)

    class ProgressMixin(Infinite):
        VERBOSE = True

        def reset(self, message: str) -> None:
            self.index = 0
            self.message = f'{message} '

        def next(self, *args, **kwargs):
            if not self.VERBOSE:
                return
            return super().next(*args, **kwargs)

        def finish(self):
            if not self.VERBOSE:
                return
            return super().finish()

        @classmethod
        def set_verbose(cls, verbose):
            cls.VERBOSE = verbose

    class Counter(counter.Counter, ProgressMixin):
        ...

    class Bar(bar.IncrementalBar, ProgressMixin):
        def reset(self, message: str, max: int | float, item: str = "") -> None:
            self.max = max
            self.suffix = '%(index)d/%(max)d ' + item
            super().reset(message)

    def __init__(self) -> None:
        self.reset = '\033[0m' + Console.rgb(231, 251, 255)
        self.bold = '\033[1m'
        self.italic = '\033[3m'

        self.bar = Console.Bar()
        self.counter = Console.Counter()

        # assign colors to instance as attributes
        for i, c in enumerate([
            (61, 187, 255),  # c1: blue
            (102, 255, 150),  # c2: green
            (126, 229, 252),  # c3: light blue
            (255, 216, 125),  # c4: yellow
            (255, 107, 66),  # c5: red
        ], 1):
            setattr(self, f'c{i}', Console.rgb(*c))

        self.danger = Console.rgb(255, 71, 77)  # red

    def __getattribute__(self, __name: str) -> Any:
        if not Console.CALLED and Console.VERBOSE:
            Console.CALLED = True
            Console.print_header()
        return object.__getattribute__(self, __name)

    @classmethod
    def set_verbose(cls, verbose: bool) -> None:
        cls.VERBOSE = verbose
        cls.Log.set_verbose(verbose)
        cls.Counter.set_verbose(verbose)
        cls.Bar.set_verbose(verbose)

    def log(self, text: str) -> Log:
        return self.Log(f'{text}{self.reset}')

    def reset_bar(self, message: str, max: int | float, item: str) -> None:
        self.bar.reset(self.log_subprocess(message), max, item)

    def reset_counter(self, message: str) -> None:
        self.counter.reset(self.log_subprocess(message))

    def elapsed_time(self, st: float | int) -> Log:
        return self.log(
            f'\t\N{stopwatch} {self.c4}{self.italic}Elapsed time: {self.bold}{self.c5}{round((time()-st) * 100) / 100}s\n')

    def log_process(self, text: str) -> Log:
        return self.log(f'{self.bold}{self.c1}{text}')

    def log_disk_op(self, object_name: str, filename: str, read: bool = False) -> Log:
        return self.log_process(
            f'\N{floppy disk} {"Reading" if read else "Writing"} {self.c2}{object_name.upper()}{self.c1} {"from" if read else "to"} disk: {self.c2}{self.bold}{filename}{self.c1}...')

    def log_subprocess(self, text: str) -> Log:
        return self.log(f'\t{self.c3}{text}')

    def error(self, error_class: Exception, text: str) -> None:
        raise error_class(self.log(f'\n\t\N{skull} {self.danger}{text}\n'))

    def warn(self, text: str) -> None:
        print(f"\n\N{cross mark} {self.danger}Warning: {text}\n")

    @staticmethod
    def rgb(r, g, b):
        return f'\u001b[38;2;{r};{g};{b}m'

    @classmethod
    def print_header(cls):
        spectrum = np.array([
            (246, 90, 62),
            (249, 164, 69),
            (240, 232, 72),
            (130, 230, 96),
            (52, 202, 197),
            (141, 126, 179),
            (246, 90, 62),
        ])

        gamut = "~ GAMuT: Granular Audio Musaicing Toolkit ~"
        spectrum = [cls.rgb(*col) for col in np.array([resample_array(val, len(gamut))
                                                       for val in spectrum.T]).T.astype('int32')]

        print("".join([f'{col}\033[1m{char}' for char, col in zip(gamut, spectrum)]) + "\n")



from __future__ import annotations
import sounddevice as sd
import numpy as np
from librosa import load
from soundfile import write
from os.path import splitext, basename
from scipy import signal
from typing_extensions import Self
from collections.abc import Iterable

from .config import AUDIO_FORMATS, CONSOLE, get_elapsed_time
from .utils import resample_array
from .controls import Envelope
from .controls import Points
from . import catch_keyboard_interrupt


class AudioBuffer:
    """
    Audio buffer class to read, write, and play back audio files.

    y: np.array
        numpy array of audio samples.

    sr: int | None
        audio sampling rate.

    bit_depth: int
        audio bit depth.
    """

    def __init__(self, y: np.ndarray | None = None, sr: int | None = None, bit_depth: int = 24) -> None:
        self.y = y
        self.sr = sr
        self.bit_depth = bit_depth

    @catch_keyboard_interrupt(lambda: CONSOLE.log_process("\N{speaker with cancellation stroke} Audio stopped").print())
    def play(self, blocking: bool = True) -> None:
        """ Plays back inner audio buffer """
        CONSOLE.log_process('\N{speaker}Playing audio...').print()
        try:
            sd.play(self.y, samplerate=self.sr, blocking=blocking)
        except sd.PortAudioError:
            CONSOLE.error(
                sd.PortAudioError,
                f'Unable to play audio stream. It\'s possible the number of output channels in the audio source ({self.chans}) is greater than what your device supports.')

    def stop(self) -> None:
        """ Stops audio buffer when using ``blocking=True`` in play method """
        sd.stop()

    def set_sampling_rate(self, sr: int) -> None:
        """ Sampling rate setter method """
        if self.sr != sr:
            self.y = resample_array(self.y, int(round(len(self.y) * sr/self.sr)))
        self.sr = sr

    def read(self, input_dir: str, sr: int | None = None, mono: bool = False) -> Self:
        """ Reads a ``.wav`` or ``.aif`` audio file from disk """
        self.y, self.sr = load(input_dir, sr=sr, mono=mono)
        self.y = self.y.T if len(self.y.shape) > 1 else self.y[:, np.newaxis]
        return self

    @property
    def chans(self) -> int:
        return self.y.shape[1]

    @property
    def samps(self) -> int:
        return self.y.shape[0]

    @get_elapsed_time
    def write(self, output_dir: str) -> None:
        """Writes audio to disk. This function is a simple wrapper of `sf.write()`"""
        ext = splitext(output_dir)[1]
        if ext not in AUDIO_FORMATS:
            CONSOLE.error(ValueError, f'Output file format must be one of the following: {AUDIO_FORMATS}')
        CONSOLE.log_disk_op('audio file', basename(output_dir)).print()
        write(output_dir, self.y, self.sr, f'PCM_{self.bit_depth}')

    def to_mono(self) -> None:
        """ Converts audio channels to mono """
        if self.chans == 1:
            return
        self.y = (self.y.sum(axis=1)[: np.newaxis] / self.chans).reshape((self.samps, 1))

    def convolve(self, impulse_response: Self | str, mix: int | float | Iterable | Envelope = 0.125, normalize: bool = True) -> None:
        """ Applies impulse response convolution to audio """
        def parse_mix_param(mix, N):
            if isinstance(mix, Envelope):
                mix_param = mix.get_points(N)
            elif isinstance(mix, Iterable):
                mix_param = Envelope(mix).get_points(N)
            else:
                mix_param = Points().fill(N, mix)
            return mix_param[:, np.newaxis]

        y_samps = self.samps

        # prepare impulse response and generate convolved signal
        ir = impulse_response if isinstance(impulse_response, AudioBuffer) else AudioBuffer().read(impulse_response, mono=True)
        ir.to_mono()
        y_wet = signal.convolve(in1=self.y, in2=ir.y, method='auto')

        # reshape to match convolved signal
        y_dry = np.zeros(shape=y_wet.shape)
        y_dry[:self.samps] = self.y
        self.y = y_dry

        # normalize
        if normalize:
            y_wet = (y_wet / np.max(np.abs(y_wet))) * np.sqrt(0.5)

        # parse wet/dry mix control parameter
        mix_param = np.zeros(shape=(y_wet.shape[0], 1))
        tmp = parse_mix_param(mix, y_samps)
        mix_param[:y_samps] = tmp
        mix_param[y_samps:] = tmp[-1]

        # mix both signals
        self.y = y_dry * (1 - mix_param) + y_wet * mix_param



# typing
from __future__ import annotations

# misc
from pathlib import Path
import os
import json


GAMUT_FILES_DIRECTORY = os.path.join(Path.home(), '.gamut')
CORPUS_DIR = os.path.join(GAMUT_FILES_DIRECTORY, 'corpora')
MOSAIC_DIR = os.path.join(GAMUT_FILES_DIRECTORY, 'mosaics')
SESSION_DATA_FILE = os.path.join(GAMUT_FILES_DIRECTORY, 'session_data.json')
CORPUS_CACHE = {}
MOSAIC_CACHE = {}


def create_root_directories() -> None:
    """ Create data storage directories """
    for directory in [GAMUT_FILES_DIRECTORY, CORPUS_DIR, MOSAIC_DIR]:
        if not os.path.exists(directory):
            os.mkdir(directory)


create_root_directories()


class GamutSession:

    def __init__(self) -> None:
        self.last_dir = None
        self.load()

    def clean_last_dir(self, attr: str, value: str | float | int) -> str | float | int:
        if os.path.exists(value):
            return value
        default = self.get_default_attrs()[attr]
        setattr(self, attr, default)
        return default

    def get_default_attrs(self) -> dict:
        return {
            'last_dir': str(Path.home()),
        }

    def load(self) -> None:
        # create session file if it doesn't exist
        if not os.path.exists(SESSION_DATA_FILE):
            with open(SESSION_DATA_FILE, 'w') as f:
                json.dump(self.get_default_attrs(), f)

        # open file an populate attributes
        with open(SESSION_DATA_FILE, 'rb') as f:
            data = json.load(f)
            default_attrs = self.get_default_attrs()
            for attr in vars(self):
                setattr(self, attr, data[attr] if attr in data else default_attrs[attr])

    def save(self) -> None:
        with open(SESSION_DATA_FILE, 'w') as f:
            json.dump(vars(self), f)

    def get(self, attr: str) -> str | float | int:
        value = getattr(self, attr)
        cleaner = getattr(self, f"clean_{attr}")
        if cleaner:
            clean_value = cleaner(attr, value)
            if clean_value != value:
                self.set(attr, clean_value)
            return clean_value
        return value

    def set(self, attr: str, value: str | float | int) -> str | float | int:
        setattr(self, attr, value)
        self.save()


GAMUT_SESSION = GamutSession()
